# File: ventaxiaiot/commands.py
import asyncio
import json
import logging

from ventaxiaiot.pending_request_tracker import PendingRequestTracker
from ventaxiaiot.sentinel_kinetic import SentinelKinetic

_LOGGER = logging.getLogger(__name__)

class VentClientCommands:
    def __init__(self, wifi_device_id, pendingRequestTracker : PendingRequestTracker):
        self._msg_id = 1
        self.device = SentinelKinetic()
        self.wifi_device_id = wifi_device_id
        self.tracker = pendingRequestTracker
        _LOGGER.debug("VentClientCommands __init__ called")

    def _next_msg_id(self):
        val = self._msg_id
        self._msg_id += 1
        return val

    async def send_subscribe(self, client):
        topics = [
            ("sub", "rd"),
            ("sub", "ee"),
            ("get", "rd"),
            ("pub", "wr", {"tsreq": 1}),
            ("get", "ee")           
        ] 

        for mtype, t_suffix, *extra in topics:
            msg = {"m": mtype, "i": self._next_msg_id()}
            if t_suffix:
                msg["t"] = f"{self.wifi_device_id}/{t_suffix}"
            if mtype == "pub":
                msg["d"] = extra[0]
                msg["f"] = 4           
            await client.send(json.dumps(msg))
            await asyncio.sleep(0.1)

    async def send_cfg_command(self, client, cmd: str):
        msg_id = self._next_msg_id()
        msg = {
            "m": "cfg",
            "cfgcmd": cmd,
            "i": msg_id
        }
        self.tracker.add(msg_id, {"cfgcmd": cmd})
        await client.send(json.dumps(msg))
        await asyncio.sleep(0.1)

    async def send_boost_request(self, client):
        msg = {
            "m": "pub",
            "t": f"{self.wifi_device_id}/wr",
            "d": {"ar_af": 3, "ar_min": 15},
            "f": 4,
            "i": self._next_msg_id(),
        }
        await client.send(json.dumps(msg))
        await asyncio.sleep(0.1)
        
    async def send_airflow_mode_request(self, client, mode: str, duration: int):
        if mode not in self.device.AIRFLOW_MODES:
            raise ValueError(f"Invalid mode: {mode}. Must be one of {list(self.device.AIRFLOW_MODES.keys())}")
        if duration not in self.device.VALID_DURATIONS:
            raise ValueError(f"Invalid duration: {duration}. Must be one of {sorted(self.device.VALID_DURATIONS)}")

        mode_code = self.device.AIRFLOW_MODES[mode]
        msg = {
            "m": "pub",
            "t": f"{self.wifi_device_id}/wr",
            "d": {
                "ar_af": mode_code,
                "ar_min": duration,
            },
            "f": 4,
            "i": self._next_msg_id(),
        }
        await client.send(json.dumps(msg))
        await asyncio.sleep(0.1)


# File: ventaxiaiot/messages.py
import json
import logging

from ventaxiaiot.pending_request_tracker import PendingRequestTracker
from ventaxiaiot.sentinel_kinetic import SentinelKinetic

_LOGGER = logging.getLogger(__name__)

class VentMessageProcessor:
    def __init__(self, pendingRequestTracker : PendingRequestTracker):
        self.device = SentinelKinetic()
        self.tracker = pendingRequestTracker
        
    async def process(self, raw_message):
        try:
            data = json.loads(raw_message)
        except json.JSONDecodeError:
            return


        self.device.apply_payload(data, self.tracker)

        if self._is_ack(data):
            await self._handle_ack(data)
        elif data.get("m") == "pub":
            await self._handle_publish(data)
        elif "r" in data:
            await self._handle_response(data)
        else:
            await self._handle_unknown(data)
            
            
        # Print the updated device state after processing each message
        _LOGGER.debug(self.device)  # concise
        # OR, for full dict print:
        # print(asdict(self.device))            


    def _is_ack(self, data):
        return isinstance(data, dict) and len(data) == 1 and "i" in data

    async def _handle_ack(self, data):
        _LOGGER.debug(f"ACK: {data}")

    async def _handle_publish(self, data):
        _LOGGER.debug(f"Publish: {data.get('t')}, Data: {data.get('d')}")

    async def _handle_response(self, data):
        _LOGGER.debug(f"Response: {data}")

    async def _handle_unknown(self, data):
        _LOGGER.debug(f"Unknown: {data}")
        





# File: ventaxiaiot/client.py
import asyncio
import ssl
import sys
import logging

_LOGGER = logging.getLogger(__name__)

has_native_psk = sys.version_info >= (3, 13)

if not has_native_psk:
    raise RuntimeError("Native PSK requires Python 3.13+")

class AsyncNativePskClient:
    def __init__(self, wifi_device_id, identity, psk_key, host, port, loop=None):
        self.identity = identity
        self.psk_key = psk_key.encode('utf-8')
        self.host = host
        self.port = port
        self.wifi_device_id = wifi_device_id
        self.loop = loop or asyncio.get_event_loop()
        self.reader = None
        self.writer = None
        self._message_queue = asyncio.Queue()
        self._reader_task = None
        self._running = False

    def psk_callback(self, hint):
        return self.identity, self.psk_key
    
    def __aiter__(self):
        return self
    
    async def __anext__(self):
        if not self._running:
            raise StopAsyncIteration
        try:
            msg = await self._message_queue.get()
            return msg
        except asyncio.CancelledError:
            raise StopAsyncIteration

    async def connect(self, timeout=30.0):
        context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
        context.set_ciphers("PSK-AES128-CBC-SHA")
        context.check_hostname = False
        context.verify_mode = ssl.CERT_NONE
        context.set_psk_client_callback(self.psk_callback)

        try:
            self.reader, self.writer = await asyncio.wait_for(
                asyncio.open_connection(self.host, self.port, ssl=context),
                timeout=timeout,
            )
            self._running = True
            self._reader_task = asyncio.create_task(self._reader_loop())
        except Exception as e:
            raise ConnectionError(f"Failed to connect: {e}")

    async def send(self, message):
        if self.writer is None:
            raise ConnectionError("Not connected")
        self.writer.write((message + '\0').encode('utf-8'))
        await self.writer.drain()      
       
    async def _reader_loop(self):
        buffer = b""
        try:
            while not self.reader.at_eof(): # type: ignore
                chunk = await self.reader.read(1024) # type: ignore
                if not chunk:
                    break
                buffer += chunk
                while b'\0' in buffer:
                    msg_bytes, buffer = buffer.split(b'\0', 1)
                    msg = msg_bytes.decode('utf-8').strip()
                    await self._message_queue.put(msg)
        except asyncio.CancelledError:
            pass
        except Exception as e:
            _LOGGER.error(f"Reader loop error: {e}")   
    
    async def close(self):
        self._running = False

        # Cancel the reader task if running
        if self._reader_task:
            self._reader_task.cancel()
            try:
                await self._reader_task
            except asyncio.CancelledError:
                pass

        # Try to close the TLS connection, suppress bad shutdown behavior
        if self.writer:
            try:
                self.writer.close()
                await self.writer.wait_closed()
            except ssl.SSLError as e:
                if "application data after close notify" in str(e):
                    _LOGGER.debug(" Ignored non-compliant TLS shutdown: ", e)
                else:
                    raise


# File: ventaxiaiot/sentinel_kinetic.py

from dataclasses import dataclass, field
import json
from typing import Any, List, Optional
from datetime import datetime
import logging

from ventaxiaiot.pending_request_tracker import PendingRequestTracker

_LOGGER = logging.getLogger(__name__)

@dataclass
class SentinelKinetic:
    # Metadata
    m: Optional[str] = None  # message type
    t: Optional[str] = None  # topic
    i: Optional[int] = None  # message ID
    
    # Core operational state
    ar_af: Optional[int] = None  # User Airflow mode: 1=low, 2=normal, 3=boost, 4=purge 
    ar_min: Optional[int] = None  # Boost duration
    ts: Optional[int] = None  # Device timestamp (UNIX)

    # Environmental sensors
    temp_indoor: Optional[float] = None
    temp_outdoor: Optional[float] = None
    humidity: Optional[float] = None

    # Config / network status
    netmode: Optional[str] = None
    wifi_signal: Optional[int] = None

    # Device flags / identifiers
    serial: Optional[str] = None
    model: Optional[str] = None
    firmware_version: Optional[str] = None

    # Raw message cache (for optional access)
    # last_update: datetime = field(default_factory=datetime.utcnow)
    # last_raw: Dict[str, Any] = field(default_factory=dict)  
    
    # Realtime status
    set_rtc: Optional[str] = None
    cm_af_sup: Optional[int] = None # supply air  flow
    cm_af_exh: Optional[int] = None # exhust  air  flow
    hand: Optional[int] = None

    # Airflow stages
    af1sup: Optional[int] = None
    af1exh: Optional[int] = None
    af2sup: Optional[int] = None
    af2exh: Optional[int] = None
    af3sup: Optional[int] = None
    af3exh: Optional[int] = None
    af4sup: Optional[int] = None
    af4exh: Optional[int] = None

    # Identification
    dname: Optional[str] = None
    svcpn: Optional[str] = None
    loc: Optional[str] = None

    # Airflow settings
    af_ls1: Optional[int] = None
    af_sw1: Optional[int] = None
    af_p1: Optional[int] = None
    af_p2: Optional[int] = None
    af_irh: Optional[int] = None
    af_enp: Optional[int] = None
    af_do: Optional[int] = None

    # Logic switches and timers
    ls1o: Optional[int] = None
    ls1d: Optional[int] = None
    sw1t: Optional[int] = None
    p1t: Optional[int] = None
    p2t: Optional[int] = None

    # Preset thresholds
    p1a_lo: Optional[int] = None
    p1a_hi: Optional[int] = None
    p1b_lo: Optional[int] = None
    p1b_hi: Optional[int] = None
    p1c_lo: Optional[int] = None
    p1c_hi: Optional[int] = None

    p2a_lo: Optional[int] = None
    p2a_hi: Optional[int] = None
    p2b_lo: Optional[int] = None
    p2b_hi: Optional[int] = None
    p2c_lo: Optional[int] = None
    p2c_hi: Optional[int] = None

    irh_sp: Optional[int] = None

    # Device mode and runtime
    m_ctl: Optional[int] = None
    m_afr: Optional[int] = None
    m_byp: Optional[int] = None
    by_ti: Optional[int] = None
    by_to: Optional[int] = None
    shrs: Optional[int] = None
    flife: Optional[int] = None

    # Settings & Flags
    btn_en: Optional[int] = None
    moda: Optional[int] = None
    ctry: Optional[int] = None
    cfgf: Optional[int] = None
    svci: Optional[int] = None

    # Relay flags
    rlf1: Optional[int] = None
    rlf2: Optional[int] = None
    rlf1sp: Optional[int] = None
    rlf2sp: Optional[int] = None
    
        # Firmware / Info
    swver: Optional[int] = None
    wfver: Optional[int] = None
    mser: Optional[int] = None
    rtc: Optional[str] = None
    su_ver: Optional[int] = None

    # Runtime & diagnostics
    runh: Optional[int] = None
    shrem: Optional[int] = None
    fhrem: Optional[int] = None
    bfit: Optional[int] = None
    fl_l: Optional[int] = None
    fl_r: Optional[int] = None

    # Airflow/Temp/Humidity Sensors
    exr_rh: Optional[int] = None
    itk_rh: Optional[int] = None
    exr_t: Optional[int] = None
    itk_t: Optional[int] = None
    sup_t: Optional[int] = None
    exh_t: Optional[int] = None
    exr_thm: Optional[int] = None
    itk_thm: Optional[int] = None
    exr_sen: Optional[int] = None
    itk_sen: Optional[int] = None
    exr_f: Optional[int] = None
    itk_f: Optional[int] = None
    exr_ft: Optional[int] = None
    itk_ft: Optional[int] = None
    exr_cp: Optional[int] = None
    itk_cp: Optional[int] = None

    sup_pwm: Optional[int] = None
    exh_pwm: Optional[int] = None
    sup_rpm: Optional[int] = None  # supply rpm
    exh_rpm: Optional[int] = None  # exhust rpm 

    # Control & status
    mstat: Optional[int] = None
    byp: Optional[int] = None
    mflags: Optional[int] = None
    mdl_afu: Optional[int] = None

    # Auto-settings
    as_af: Optional[int] = None  # User Airflow mode: 1=low, 2=normal, 3=boost, 4=purge  
    as_ts: Optional[str] = None
    as_rsec: Optional[int] = None 
    as_oa: Optional[int] = None
    as_os: Optional[int] = None

    # Power
    pwr: Optional[int] = None #Power as a %
    afr_m: Optional[int] = None
    afr_php: Optional[int] = None

    # Server
    svr_conn: Optional[int] = None
    svrerr: Optional[int] = None
    svrstat: Optional[int] = None
    cmsrc: Optional[int] = None

    # Mode names
    mn: List[Optional[str]] = field(default_factory=lambda: [None]*16)
    
    AIRFLOW_MODES = {
        "reset": 0, # This stops the current user requested airflow mode
        "normal": 2,
        "boost": 3,
        "purge": 4,
    }
    
    VALID_DURATIONS = {0,15, 30, 45, 60} # 0 is sent with "reset": 0 to stop the current user requested airflow mode

    def apply_payload(self, data: Any,tracker : PendingRequestTracker):
        """Initialize or update the instance with nested or flat payload, handling mn0–mn15."""
        try:
            # Step 0: If input is a raw JSON string, try to parse it
            if isinstance(data, str):
                data = json.loads(data)
                
            # Step 0.5: Handle response metadata and assign to self
            msg_id = data.get("i")
            if msg_id is not None:
                meta = tracker.pop(msg_id) 
                if meta:
                    _LOGGER.debug(f"✅ Matched metadata for msg_id {msg_id}: {meta}")
                    key = meta.get("cfgcmd")
                    value = data.get("r")
                    if key and value is not None:
                        # Clean up key (e.g., "netmode?" → "netmode")
                        clean_key = key.rstrip("?")
                        _LOGGER.debug(f"🔧 Setting attribute: {clean_key} = {value}")
                        setattr(self, clean_key, value)    
                        return  # ✅ Done — no need to process further
                

            # Step 1: Unwrap nested keys if present
            if isinstance(data, dict):
                flat_data = data.get('d') or data.get('r') or data
            else:
                raise ValueError(f"Unexpected data type: {type(data)}")
            
            # Step 2.1: Define fields to ignore (e.g., internal timestamps)
            IGNORED_FIELDS = {f"ts{i}" for i in range(1, 28)}
            IGNORED_FIELDS.update({
                "mlang", "diag", "ls", "sw", "ps1", "ps2", "dsav", "warn", "secpin",
                "byp_ept", "afmax", "afcnt", "f90", "brid"
            })    
        

            # Step 2: Extract mn0–mn15 into a list and assign to `mn`
            mn_list = [flat_data.get(f'mn{i}') for i in range(16)]
            if any(v is not None for v in mn_list):
                self.mn = mn_list

            # Step 3: Update all fields defined in the dataclass
            for key, value in flat_data.items():
                if key.startswith("mn") and key[2:].isdigit(): 
                    continue  # already handled
                if hasattr(self, key):
                    setattr(self, key, value)
                elif key not in IGNORED_FIELDS:
                    _LOGGER.debug(f"⚠️ Unknown field ignored: {key}")

        except Exception as e:
            _LOGGER.error(f"❌ apply_payload failed: {e}")

    @property
    def extract_temp_c(self) -> Optional[float]:
        """Extract (indoor) temperature in Celsius."""
        return self.exr_t / 10 if self.exr_t is not None and self.exr_t > -1000 else None
      
    @property
    def outdoor_temp_c(self) -> Optional[float]:
        """Outdoor (external) temperature in Celsius."""
        return self.itk_t / 10 if self.itk_t is not None and self.itk_t > -1000 else None   

    @property
    def supply_temp_c(self) -> Optional[float]:
        """Supply air temperature in Celsius."""           
        # Check if there's another field for supply temp; if not, maybe None
        if self.sup_t is not None and self.sup_t != -1000:
            return self.sup_t / 10.0
        return None
    
    @property
    def summer_bypass_indoor_temp(self) -> Optional[float]:
        """Summer bypass indoor temperature in Celsius."""           
        # Check if there's another field for supply temp; if not, maybe None
        return self.by_ti / 10 if self.by_ti is not None and self.by_ti > -1000 else None    
    
    @property
    def summer_bypass_outdoor_tempmp_c(self) -> Optional[float]:
        """Summer bypass outdoor temperature in Celsius."""           
        return self.by_to / 10 if self.by_to is not None and self.by_to > -1000 else None  
       
    
    @property
    def get_user_airflow_mode(self) -> Optional[str]:
        """Translate airflow setting to a friendly label"""
        user_airflow_modes = {
            1: "Normal",
            2: "Low",
            3: "Boost",
            4: "Purge"
        }
        return user_airflow_modes.get(self.as_af) if self.as_af is not None else "Unknown" # type: ignore

    def __str__(self):
        return f"<VentAxiaDevice name={self.dname} airflow_mode={self.get_user_airflow_mode} supply={self.sup_rpm} exhaust={self.exh_rpm}>"



# File: ventaxiaiot/cli.py
import asyncio
import argparse
import json
import os
import ssl
import sys
import logging

from ventaxiaiot.client import AsyncNativePskClient
from ventaxiaiot.messages import VentMessageProcessor
from ventaxiaiot.commands import VentClientCommands
from ventaxiaiot.pending_request_tracker import PendingRequestTracker

_LOGGER = logging.getLogger(__name__)

def load_config(path=None):
    if not path:
        path = "config.json"

    if os.path.exists(path):
        with open(path, "r") as f:
            return json.load(f)

    return {}

def setup_cli_logging(level: str = "INFO"):
    logging.basicConfig(
        level=getattr(logging, level.upper(), logging.INFO),
        format="%(levelname)s: %(message)s"
    )

def merge_config(cli_args, file_config):
    merged = file_config.copy()
    for key in vars(cli_args):
        value = getattr(cli_args, key)
        if value is not None:
            merged[key] = value
    return merged


def build_parser():
    parser = argparse.ArgumentParser(description="Vent-Axia IoT command-line interface")
    parser.add_argument("--config", default="config.json", help="Path to configuration file")
    
    parser.add_argument(
        "--log-level",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        default="INFO",
        help="Logging level (default: INFO)"
    )
       

    subparsers = parser.add_subparsers(dest="command", required=True)

    # Example: status command
    status_cmd = subparsers.add_parser("status", help="Query device network mode")
    status_cmd.add_argument("--host", help="Device host/IP")
    status_cmd.add_argument("--port", type=int, help="Port number")
    status_cmd.add_argument("--identity", help="Device identity string")
    status_cmd.add_argument("--psk_key", help="Pre-shared key")
    status_cmd.add_argument("--wifi_device_id", help="WiFi device ID")

    #Will add more subcommands here, like `subscribe`, `set`, `get`, etc.
    return parser


async def run_status(config):
    tracker = PendingRequestTracker()
    client = AsyncNativePskClient(
        config["wifi_device_id"],
        config["identity"],
        config["psk_key"],
        config["host"],
        config["port"]
    )

    processor = VentMessageProcessor(tracker)
    commands = VentClientCommands(client.wifi_device_id, tracker)

    await client.connect()
    await commands.send_subscribe(client)    
    await commands.send_cfg_command(client, "netmode?")

    try:
        async for msg in client:
            await processor.process(msg)
    except KeyboardInterrupt:
        _LOGGER.debug("⏹️ Interrupted by user")
    except Exception as e:
        _LOGGER.error("Error processing message: %s", e)
    finally:
        await client.close()


async def main(): 
    
    if len(sys.argv) == 1:
        sys.argv.append("status")  # Inject default command if none given
    
    parser = build_parser()
    cli_args = parser.parse_args()

    file_config = load_config(cli_args.config)
    config = merge_config(cli_args, file_config)
    setup_cli_logging(cli_args.log_level)

    # Validate required fields
    required_keys = ["host", "port", "identity", "psk_key", "wifi_device_id"]
    for key in required_keys:
        if key not in config:
            raise ValueError(f"Missing required config value: {key}")

    if cli_args.command == "status":
        await run_status(config)       